"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const Node_1 = __importDefault(require("./Node"));
/*
 * Data structure of nodes that contains a head, tail
 * and length property.
 *
 * Each node has value and a next pointer to another node, or null.
 *
 * Unlike an array:
 * - It does not have an index
 * - Each node is only linked to the next node
 * - Random access is not allowed
 *
 * Good for fast insertion and deletion of data.
 * Bad for accessing data.
 */
class SinglyLinkedList {
    constructor() {
        this.head = null;
        this.tail = null;
        this.length = 0;
    }
    push(val) {
        const newNode = new Node_1.default(val);
        if (!this.head) {
            this.head = newNode;
            this.tail = this.head;
        }
        else {
            if (this.tail !== null) {
                this.tail.next = newNode;
                this.tail = newNode;
            }
        }
        this.length++;
        return this;
    }
    pop() {
        if (!this.head)
            return undefined;
        let current = this.head;
        let newTail = current;
        while (current.next) {
            newTail = current;
            current = current.next;
        }
        this.tail = newTail;
        this.tail.next = null;
        this.length--;
        if (this.length === 0) {
            this.head = null;
            this.tail = null;
        }
        return current;
    }
    shift() {
        if (!this.head)
            return undefined;
        let removedHead = this.head;
        this.head = removedHead.next;
        this.length--;
        if (this.length === 0) {
            this.tail = null;
        }
        return removedHead;
    }
    unshift(val) {
        const newNode = new Node_1.default(val);
        if (!this.head) {
            this.head = newNode;
            this.tail = this.head;
        }
        else {
            newNode.next = this.head;
            this.head = newNode;
        }
        this.length++;
        return this;
    }
    get(n) {
        if (n < 0 || n >= this.length)
            return undefined;
        let i = 0;
        let current = this.head;
        while (i !== n) {
            current = current === null || current === void 0 ? void 0 : current.next;
            i++;
        }
        return current;
    }
    set(n, val) {
        const foundNode = this.get(n);
        if (foundNode !== null && foundNode !== undefined) {
            foundNode.val = val;
            return true;
        }
        return false;
    }
    insert(n, val) {
        if (n < 0 || n > this.length)
            return false;
        if (n === 0)
            return Boolean(this.unshift(val));
        if (n === this.length)
            return Boolean(this.push(val));
        const prevNode = this.get(n - 1);
        if (prevNode) {
            const newNode = new Node_1.default(val);
            const temp = prevNode === null || prevNode === void 0 ? void 0 : prevNode.next;
            prevNode.next = newNode;
            newNode.next = temp;
            this.length++;
            return true;
        }
        return false;
    }
    remove(n) {
        if (n < 0 || n >= this.length)
            return undefined;
        if (n === 0)
            return this.shift();
        if (n === this.length - 1)
            return this.pop();
        const prevNode = this.get(n - 1);
        if (prevNode && prevNode.next) {
            const removed = prevNode === null || prevNode === void 0 ? void 0 : prevNode.next;
            prevNode.next = removed === null || removed === void 0 ? void 0 : removed.next;
            this.length--;
            return removed;
        }
    }
    reverse() {
        // swap head and tail
        let node = this.head;
        this.head = this.tail;
        this.tail = node;
        let prev = null;
        let next = null;
        // swap all nodes
        for (let i = 0; i < this.length; i++) {
            next = node === null || node === void 0 ? void 0 : node.next;
            if (node) {
                node.next = prev;
                prev = node;
            }
            if (next) {
                node = next;
            }
        }
        return this;
    }
}
exports.default = SinglyLinkedList;
